/* rsg.c - Random Sentence Generator, Alain Brobecker 2009/01/07 */
/* Uses context free grammar to produce random sentences */
#include <stdio.h>  /* file functions, etc... */
#include <stdlib.h> /* exit(), malloc() */
#include <time.h>

#define MaxSentenceLength 99999

/**** From D.E.Knuth's Art of Computer Programming, vol.2, p185 ****/
/**** period of pseudorandom sequence is 2147483646             ****/
#define rndM 2147483647 /* 2^31-1, a Mersenne prime */
#define rndA 48271      /* this does well in the spectral test */
#define rndQ 44488      /* rndM/rndA */
#define rndR 3399       /* rndM%rndA, it's important that rndR<rndQ */
long rndX;              /* should be initialised to 0<rndX<rndM */
void rndNext() {
  rndX=rndA*(rndX%rndQ)-rndR*(long)(rndX/rndQ); /* X=AX mod M */
  if(rndX<0) {rndX+=rndM; }
}          

/**** PrintString **********************************************************/
/* Just because i don't want to use the null terminator in account */
void PrintString(char *String,unsigned int StringLength) {
  int c;
  for(c=0;c<StringLength;c++) {
    printf("%c",String[c]);
  }
}

/**** SearchNext routine, brute force version ******************************/
unsigned int SearchNext(char *String,unsigned int StringLength,
               char *Text,unsigned int TextLength,int TextPos) {
  int NbMatch=0;
  while((NbMatch<StringLength) && (TextPos+StringLength<=TextLength)) {
    if(Text[TextPos+NbMatch]==String[NbMatch]) {
      NbMatch++;
    } else {
      NbMatch=0;
      TextPos++;
    }
  }
  if(NbMatch==StringLength) {
    return(TextPos);
  } else {
    return(-1);
  }
}

/**** Displace routine *****************************************************/
/* Returns new length (or quit if it would exceed MaxLength). There are    */
/* two displacement routines depending on the */
int Displace(char *Text,unsigned int Length,int MaxLength,int OldPos,int NewPos) {
  int c;
  if(NewPos<OldPos) {
    for(c=0;c<=Length-1-OldPos;c++) {
      Text[NewPos+c]=Text[OldPos+c];
    }
  } else if(OldPos<NewPos) {
    if(Length+NewPos-OldPos>MaxLength) {
      printf("error: not enough memory to displace string\n");
      exit(1);
    }
    for(c=Length-1-OldPos;c>=0;c--) {
      Text[NewPos+c]=Text[OldPos+c];
    }
  }
  return(Length+NewPos-OldPos);
}

/**** Main *****************************************************************/
int main(argc,argv) int argc; char *argv[]; {
  char *Grammar,*Sentence;
  unsigned int GrammarLength,SentenceLength;
  unsigned int VarStart,VarEnd,VarLength;
  unsigned int ChoiceStart,ChoiceEnd,ChoiceLength,NbChoices,Choice;
  unsigned int c;
  FILE *filep;

  setbuf(stdout, NULL); /* no buffering for output! */
  Sentence=(char *) malloc((unsigned int) MaxSentenceLength); /* memory for sentence */
  if(argc!=2) { printf("usage: rsg GrammarFile\n"); exit(1); } /* 2 args */
  rndX=1; /* initialise random seed */
  rndX=rawclock(); /* initialise random seed */
    
  /* load Grammar File and append a 'G' at start of file */
  filep=fopen(argv[1],"rb");
  if(filep==0) { printf("error: %s not found\n",argv[1]); exit(1); }
  fseek(filep,0,2); /* 2 for EOF */
  GrammarLength=(int) ftell(filep);
  fseek(filep,0,0); /* 0 for SOF */
  Grammar=(char *) malloc((unsigned int) GrammarLength+1);
  Grammar[0]='G';
  fread(Grammar+1,1,GrammarLength,filep);
  GrammarLength++;
  fclose(filep);
  
  /* Tidy up grammar file: change tabs, CR, LF into spaces. Remove double spaces and
  spaces that are after a { or before a }. Please note i don't start at position 0
  because first char can never be inside an important word. */
  for(c=0;c<GrammarLength;c++) {
    switch(Grammar[c]) {
      case 9: case 10: case 13: Grammar[c]=' '; break;
    }
    if(((Grammar[c]==' ') || (Grammar[c]=='}')) && (Grammar[c-1]==' ')) {
      GrammarLength=Displace(Grammar,GrammarLength,GrammarLength,c,c-1);
      c--; /* go back one char since we wiped one */
    }
    if((Grammar[c]==' ') && (Grammar[c-1]=='{')) {
      GrammarLength=Displace(Grammar,GrammarLength-1,GrammarLength,c+1,c)+1;
      c--; /* go back one char since we wiped one */
    }
  }
/*!!! PrintString(Grammar,GrammarLength); printf("/\n\n");*/
    
  /* Processing is mostly a search and replace loop, but with some randomness */
/*  for(c=0;c<MaxSentenceLength;c++) { Sentence[c]=0; }*/
  Sentence[0]=' '; /*start with a space so we can add a '{' */
  Sentence[1]='<';
  Sentence[2]='s'; Sentence[3]='t'; Sentence[4]='a'; Sentence[5]='r'; Sentence[6]='t';
  Sentence[7]='>';
  Sentence[8]=' ';
/*!!! Sentence=" <start> "; doesn't work due to string handling */  
  SentenceLength=9;
  VarStart=0;
  
  do{
    /* Search a non terminal variable inside remaining text */
    VarStart=SearchNext((char *) "<",1,Sentence,SentenceLength,VarStart);
    if(VarStart!=-1) {
      /* Search this non terminal variable in Grammar */
      VarEnd=SearchNext((char *) ">",1,Sentence,SentenceLength,VarStart);
      if(VarEnd==-1) {
        /*!!! VOIR LE CAS <start <toto> QUI DONNERA UNE AUTRE ERREUR */
        printf("error: non terminal variable started with < but not ended\n");
        exit(1);
      }
      VarLength=VarEnd-VarStart+1;
/*!!! printf("FOUND: "); PrintString(Sentence+VarStart,VarLength); printf(" %d\n",VarLength);*/
      c=Sentence[VarStart-1]; /* replace previous char by { in sentence */
      Sentence[VarStart-1]='{';
      ChoiceStart=SearchNext(Sentence+VarStart-1,VarLength+1,Grammar,GrammarLength,0);
      if(ChoiceStart==-1) {
        printf("error: ");
        PrintString(Sentence+VarStart,VarLength);
        printf(" not found in GrammarFile\n");
        exit(1);
      }
      Sentence[VarStart-1]=c; /* restore correct char */
      ChoiceStart=ChoiceStart+VarLength+1; /* skip the name of the variable */
      /* Count the number of possibilities for this variable by counting */
      /* the number of ; before the } */
      ChoiceEnd=SearchNext((char *) "}",1,Grammar,GrammarLength,ChoiceStart);
      if((ChoiceEnd==-1) || (Grammar[ChoiceEnd-1]!=';')) {
        printf("error: ");
        PrintString(Sentence+VarStart,VarLength);
        printf(" must end with ;}\n");
        exit(1);
      }
/*!!! PrintString(Grammar+ChoiceStart,ChoiceEnd-ChoiceStart+1); printf(" Start:%d End:%d Length: %d\n",ChoiceStart,ChoiceEnd,ChoiceEnd-ChoiceStart+1);*/
      c=ChoiceStart;
      NbChoices=-1;
      do{
        c=SearchNext((char *) ";",1,Grammar,GrammarLength,c)+1;
        NbChoices++;
      } while((c!=0) && (c<=ChoiceEnd));
      /* Choose randomly one possibility */
      /*!!! NORMALEMENT IL FAUDRAIT FAIRE UN MEILLEUR CHOIX, MAIS BON...*/
      Choice=rndX%NbChoices; /* choice in [0;NbChoices[*/
/*!!! printf("NbChoices=%d  ;  rndX=%d  ;  Choice=%d\n",NbChoices,rndX,Choice+1);*/
      rndNext();
      for(c=0;c<Choice;c++) {
        ChoiceStart=SearchNext((char *) ";",1,Grammar,GrammarLength,ChoiceStart)+1;
      }
      ChoiceEnd=SearchNext((char *) ";",1,Grammar,GrammarLength,ChoiceStart)-1;
      ChoiceLength=ChoiceEnd-ChoiceStart+1;
/*!!! PrintString(Grammar+ChoiceStart,ChoiceEnd-ChoiceStart+1); printf(" Start:%d End:%d Length: %d\n",ChoiceStart,ChoiceEnd,ChoiceEnd-ChoiceStart+1);*/
      /* Now replace non terminal variable in Sentence by possibility in ChoiceStart-ChoiceEnd */
      /* At first displace the sentence */
/*!!! printf("Length:%d instead of %d\n",ChoiceLength,VarLength);*/
      SentenceLength=Displace(Sentence,SentenceLength,MaxSentenceLength,VarEnd+1,VarStart+ChoiceLength);
      /* VarLength=VarEnd-VarStart+1;*/
      for(c=0;c<ChoiceLength;c++) {
        Sentence[VarStart+c]=Grammar[ChoiceStart+c];
      }
    }
/*!!! printf(";"); PrintString(Sentence,SentenceLength); printf("; length=%d\n",SentenceLength);*/
  } while(VarStart!=-1);
  PrintString(Sentence+1,SentenceLength-1); printf("\n");
  exit(0);
}
